
"""
Instant Photo Show - a program for creating shows from the photos taken with
                     a mobile device. In addition music may be played while
                     the photos are shown.

                     Recommendation: install miso library to keep the backlight
                     on all the time.
"""

#= Features to be added:
#= - preview while selecting photos
#= - show editing
#= - more customization: delay between photos

import os
import os.path
import sys
import e32
import appuifw
import graphics
import audio
import dir_iter
import series60_console
import pyS60uiutil
try:
    import miso
    MISO_LIB_PRESENT = True
except ImportError:
    MISO_LIB_PRESENT = False

from key_codes import \
     EKeyLeftArrow, EKeyRightArrow, EKeyUpArrow, EKeyDownArrow, EKeyDevice3

IPS_CONFIG_FILE = 'ipsconfig.txt'
IPS_DATA_DIRECTORY = 'photodat'
SHOW_CONFIG_FILE = 'show.dat'
SHOW_DIR_PREFIX = '_ips_'
SHOW_DIR_SUFFIX = '_'
VER_SHOW_FILE = 2
VER_CONF_FILE = 1

mylock = e32.Ao_lock()
global sh

def volumeControl(soundObj, style=None, bgimage=None):
    ai = pyS60uiutil.save_current_app_info()
    lock = e32.Ao_lock()
    appuifw.app.exit_key_handler = lock.signal
    appuifw.app.menu = [(u'Select', lock.signal)]
    cnvs = None
        
    def __redraw(dummy=None):
        if style == 'bar':
            cnvs.clear((0,0,0))
            max = soundObj.max_volume()
            cur = soundObj.current_volume()
            h = cnvs.size[1]
            w = cnvs.size[0]
            lo = h-20
            up = 20
            step = (lo - up) / max
            x1 = (w-50)/2
            x2 = x1+50
            cnvs.rectangle([(x1,up),(x2,lo)],
                           outline=(0x0,0xaa,0x0),
                           width=5)
            cnvs.rectangle([(x1,up+(step*(max-cur))),(x2,lo)],
                           fill=(0x11,0xCE,0x11))
        else:
            cnvs.clear()
            cnvs.text((cnvs.size[0]/2, cnvs.size[1]/2),
                      unicode(str(soundObj.current_volume())),
                      0,
                      u'Acb30')
    
    def __key_events(events):
        code = events['keycode']
        volume = soundObj.current_volume()
        if code == EKeyLeftArrow or code == EKeyDownArrow:
            volume -= 1
            if volume < 0: volume = 0
        elif code == EKeyRightArrow or code == EKeyUpArrow:
            volume += 1
            if volume > soundObj.max_volume():
                volume = soundObj.max_volume()
        soundObj.set_volume(volume)
        __redraw()

    if cnvs == None:
        cnvs = appuifw.Canvas(__redraw, __key_events)
    appuifw.app.body = cnvs    
    __redraw()
    lock.wait()
    pyS60uiutil.restore_app_info(ai)
    return soundObj.current_volume()

class disableMenuAndExitkey:
    def __init__(self):
        self.exit_key = appuifw.app.exit_key_handler
        self.menu = appuifw.app.menu
        appuifw.app.exit_key_handler = self.__nop
        appuifw.app.menu = []
    def __nop(self):
        return
    def reactivate(self):
        appuifw.app.exit_key_handler = self.exit_key
        appuifw.app.menu = self.menu
        
class Show:
    """
    Heart of the Instant Photo Show program...
    """
    def __init__(self):
        #= Detect device screen size, there are more clever ways of
        #= doing this, too! Check the manual!
        basicapp = pyS60uiutil.save_current_app_info()
        appuifw.app.screen = 'full'
        appuifw.app.body = appuifw.Canvas()
        self.screensize = appuifw.app.body.size
        pyS60uiutil.restore_app_info(basicapp)
        print 'Detected device screen size: ' + str(self.screensize)
        self.__load_parameters()
        self.reset()

    def reset(self):
        self.photo_list = []
        self.music_list = []
        self.psound = None
        self.sound_index = 0
        #self.rimg = None
        self.current_img = None
        self.border = None
        self.show_dir = None
        self.image_quality = 50

    def save_parameters(self):
        try:
            cf = file(IPS_CONFIG_FILE, 'w')
            cf.write(repr(self.config))
            cf.close()
        except:
            pass

    def __load_parameters(self):
        self.config = {}
        try:
            cf = file(IPS_CONFIG_FILE, 'r')
            self.config = eval(cf.read())
            cf.close()
        except:
            pass
        #= If the config file could not be read then create
        #= a default configuration
        self.__set_default_config()

    def __set_default_config(self):
        """
        If we had problems reading the configuration file or if the
        configuration file was an old one, set default values.
        """
        self.config.setdefault('version', VER_CONF_FILE)
        self.config.setdefault('volume', 5)
        self.config.setdefault('delay', 5)
        self.config.setdefault('photodir', None)
        self.config.setdefault('musicdir', None)
    
    def init_new_show(self, show_title, image_quality):
        self.show_dir = os.path.join(IPS_DATA_DIRECTORY,
                                     SHOW_DIR_PREFIX + \
                                     show_title + \
                                     SHOW_DIR_SUFFIX)
        if os.path.exists(self.show_dir):
            appuifw.note(u'Show already exists', 'error')
            return False
        try:
            os.mkdir(self.show_dir)
        except IOError, detail:
            appuifw.note(unicode(detail), 'error')
            return False
        self.image_quality = image_quality
        return True

    def init_existing_show(self, show_title):
        self.show_dir = os.path.join(IPS_DATA_DIRECTORY,
                                     SHOW_DIR_PREFIX + \
                                     show_title + \
                                     SHOW_DIR_SUFFIX)

    def get_title(self):
        pre_l = len(SHOW_DIR_PREFIX)
        suf_l = len(SHOW_DIR_SUFFIX)
        name = os.path.split(self.show_dir)[1]
        return unicode(name[pre_l : 0-suf_l])

    def set_title(self, show_title):
        show_title = show_title
        os.rename(self.show_dir,
                  os.path.join(IPS_DATA_DIRECTORY, SHOW_DIR_PREFIX + \
                               show_title + SHOW_DIR_SUFFIX))
        #try:
        #    show_title = show_title.decode()
        #    os.rename(self.show_dir.decode(),
        #              os.path.join(IPS_DATA_DIRECTORY, SHOW_DIR_PREFIX + \
        #                           show_title + SHOW_DIR_SUFFIX))
        #except:
        #    #= more details might be appreciated
        #    appuifw.note(u'Cannot rename show', 'error')
        #    return
        self.show_dir = os.path.join(IPS_DATA_DIRECTORY, SHOW_DIR_PREFIX + \
                                     show_title + SHOW_DIR_SUFFIX)
        for i in range(len(self.photo_list)):
            self.photo_list[i] = os.path.join(self.show_dir,
                                              os.path.split\
                                              (self.photo_list[i])[1])
        
    def delete(self, show_title):
        show_reset = False
        show_dir = os.path.join(IPS_DATA_DIRECTORY,
                                SHOW_DIR_PREFIX + \
                                show_title + \
                                SHOW_DIR_SUFFIX)
        if self.show_dir == show_dir:
            self.reset()
            show_reset = True
        for item in os.listdir(show_dir):
            try:
                os.remove(os.path.join(show_dir, item))
            except:
                type, value = sys.exc_info() [:2]
                appuifw.note(unicode(str(type)+'\n'+str(value)), 'error')
        try:
            os.rmdir(show_dir)
        except:
            type, value = sys.exc_info() [:2]
            appuifw.note(unicode(str(type)+'\n'+str(value)), 'error')
        return show_reset

    def __process_image(self, sel):
        print 'directory: ' + str(sel[0])
        print 'loading ' + str(sel[1]) + '...'
        imgparam = graphics.Image.inspect(os.path.join(sel[0], sel[1]))

        img = graphics.Image.new(imgparam['size'], 'RGB')
        print 'imgsize: ' + str(imgparam['size'])
        img.load(os.path.join(sel[0], sel[1]))
        print 'processing ' + str(sel[1]) + '...'
        img = img.transpose(graphics.ROTATE_90)
        resized_img = img.resize(self.screensize, keepaspect=1)
        print 'saving ' + str(sel[1]) + '...'
        #= perhaps 'quality level' parameter should be configurable
        resized_img.save(os.path.join(self.show_dir, sel[1]),
                         quality=self.image_quality)

    def add_photo(self, source_dir, image_name, index=None):
        self.__process_image((source_dir, image_name))
        if index == None or index >= len(self.photo_list):
            self.photo_list.append(os.path.join(self.show_dir, image_name))
        else:
            self.photo_list.insert(index,
                                   os.path.join(self.show_dir, image_name))
    
    def del_photo(self, index):
        if len(self.photo_list):
            os.remove(self.photo_list.pop(index))

    def get_photo_list(self):
        l = []
        for i in self.photo_list:
            l.append(unicode(os.path.split(i)[1]))
        if len(l): return l
        else:      return [u'<empty>']

    def set_photo_params(self, delay=None):
        if not delay == None:
            self.config['delay'] = delay
    
    def add_music(self, fullname, index=None):
        #= File not copied, i.e., it's played from its original location
        if index == None or index >= len(self.music_list):
            self.music_list.append(fullname)
        else:
            self.music_list.insert(index, fullname)
        print unicode(fullname) + u' added to play list'
            
    def del_music(self, index):
        #= Note: file not removed!
        if len(self.music_list):
            item = self.music_list.pop(index)
            print unicode(item) + u' removed from play list'

    def get_music_list(self):
        l = []
        for i in self.music_list:
            l.append(unicode(os.path.split(i)[1]))
        if len(l): return l
        else:      return [u'<empty>']

    def set_music_params(self, volume=None):
        if not volume == None:
            self.config['volume'] = volume

    def save(self):
        showdata = repr((VER_SHOW_FILE, self.image_quality,
                         self.photo_list, self.music_list))
        f = file(os.path.join(self.show_dir, SHOW_CONFIG_FILE), 'w')
        f.write(showdata)
        f.close()
    
    def load(self):
        try:
            f = file(os.path.join(self.show_dir, SHOW_CONFIG_FILE), 'r')
        except IOError, detail:
            appuifw.note(unicode(detail), 'error')
            return False
        showdata = eval(f.read())
        ver = showdata[0]
        if ver == 1:
            self.image_quality = 100
            self.photo_list = showdata[1]
            self.music_list = showdata[2]            
        else:
            self.image_quality = showdata[1]
            self.photo_list = showdata[2]
            self.music_list = showdata[3]
        return True
    
    def __play_next_music_file(self):
        if self.psound:
            self.psound.stop()
            self.psound.close()
        try:
            self.psound = audio.Sound.open(self.music_list[self.sound_index])
            self.psound.set_volume(self.config.get('volume'))
            self.psound.play()
            self.sound_index = (self.sound_index + 1 ) % len(self.music_list)
        except:
            #= Could provide a bit more info...
            appuifw.note(u'Cannot play selected music', 'error')
            self.psound = None
        
    def __redraw(self, area):
        #if self.rimg:
        #    self.canvas.blit(self.rimg)
        self.canvas.clear(0)
        if self.current_img and (not self.border == None):
            self.canvas.blit(self.current_img, target=(self.border,0))
        
    def start(self):
        basicapp = pyS60uiutil.save_current_app_info()
        sleeping_time = self.config.get('delay')

        self.stop_show = False
        self.sound_index = 0
        if len(self.music_list) > 0:
            self.__play_next_music_file()
        else:
            self.psound = None        
        self.canvas = appuifw.Canvas(self.__redraw)
        appuifw.app.body = self.canvas
        appuifw.app.screen = 'full'

        cnt = 0
        while cnt < len(self.photo_list):
            if self.stop_show == True:
                break
            self.current_img = graphics.Image.open(self.photo_list[cnt])
            #self.rimg = img.resize(self.canvas.size)
            #self.canvas.blit(self.rimg)
            if self.border == None:
                #= All the images in the show are of same size. When displaying
                #= the very first image calculate the borders so that the
                #= images can be centered to the display.
                imgsize = graphics.Image.inspect(self.photo_list[cnt])['size']
                self.border = (self.screensize[0] - imgsize[0]) / 2

            self.canvas.clear(0)
            self.canvas.blit(self.current_img, target=(self.border,0))
            cnt += 1
            if self.psound and (not self.psound.state() == audio.EPlaying):
                self.__play_next_music_file()
            if MISO_LIB_PRESENT:
                miso.reset_inactivity_time()
            e32.ao_sleep(sleeping_time)
            if MISO_LIB_PRESENT:
                miso.reset_inactivity_time()
        if self.psound:
            self.psound.stop()
            self.psound.close()
        pyS60uiutil.restore_app_info(basicapp)

    def stop(self):
        self.stop_show = True
        
#=
#= UI callbacks
#=

def playShow():
    global sh
    set_options_menu('playing')
    sh.start()
    set_options_menu('showloaded')

def stopShow():
    global sh
    sh.stop()

def createShow():
    global sh, fb

    def select_all_in_dir():
        global __break_adding_all_in_dir
        
        def select_all_exit():
            global __break_adding_all_in_dir #= would be better w/o global...
            __break_adding_all_in_dir = True

        global fb, ips_console
        currdir = fb.reveal_current_location()[0]
        #fb.release_select()

        #= Use the console while processing the images in order
        #= to allow the user know what is going on.
        fb_app_info = pyS60uiutil.save_current_app_info()

        appuifw.app.body = ips_console.text
        appuifw.app.menu = []
        appuifw.app.exit_key_handler = select_all_exit

        __break_adding_all_in_dir = False
        for item in os.listdir(currdir):
            ext = os.path.splitext(item)[1]
            if ext == '.jpg' or ext == '.jpeg' or ext == '.png':
                sh.add_photo(currdir, item)
            if __break_adding_all_in_dir: break
        if __break_adding_all_in_dir:
            appuifw.note(u'Image processing stopped', 'info')
        else:
            appuifw.note(u'All images processed', 'info')
        pyS60uiutil.restore_app_info(fb_app_info)

    show_title = appuifw.query(u'Show title', 'text')
    if show_title == None: return

    image_quality = 0
    while not ((image_quality >= 1) and (image_quality <= 100)):
        image_quality = appuifw.query(u'Image quality (1-100)', 'number', 50)
        if image_quality == None: return

    sh.reset()
    if not sh.init_new_show(show_title, image_quality):
        set_options_menu('initial')
        return

    #start_path = 'E:/Images/'
    start_path = sh.config['photodir']
    fb = pyS60uiutil.dirBrowser('Select photo',
                                [(u'All', select_all_in_dir)])

    while 1:
        sel = fb.select(start_path)
        start_path = None
        if sel == None: break
        disableCtrl = disableMenuAndExitkey()
        sh.add_photo(sel[0], sel[1])
        disableCtrl.reactivate()
        sh.config['photodir'] = sel[0]

    start_path = sh.config['musicdir']
    fb.change_ui_defs(title='Select music', menu=[])
    while 1:
        sel = fb.select(start_path)
        start_path = None
        if sel == None: break
        #= Perhaps the song should be copied to the show directory?
        sh.add_music(os.path.join(sel[0], sel[1]))
        sh.config['musicdir'] = sel[0]
    sh.save()
    set_options_menu('showloaded')

def editShow():
    global sh, ic
    tab_list = [u'General', u'Photos', u'Music']
    edit_lock = e32.Ao_lock()
    orgapp = pyS60uiutil.save_current_app_info()
    
    def tabpages(page):
        if page == 0:
            #=
            #=  More stuff later on
            #=
            def _make_lbox_menu():
                lst = [u'Title: ' + sh.get_title()]
                lst.append(u'Delay: ' + unicode(str(sh.config['delay'])))
                return lst
            def cb():
                i = lbox.current()
                if i == 0:
                    title = appuifw.query(u'Type new title', 'text',
                                          sh.get_title())
                    if not title == None:
                        sh.set_title(title)
                elif i == 1:
                    delay = appuifw.query(u'Delay between photos', 'number',
                                          sh.config['delay'])
                    if not delay == None:
                        sh.set_photo_params(delay=delay)
                #= Refresh the tabs and active page.
                appuifw.app.set_tabs(tab_list, tabpages)
                appuifw.app.activate_tab(page)
                lbox.set_list(_make_lbox_menu(), i)
            lbox = appuifw.Listbox(_make_lbox_menu(), cb)
            appuifw.app.body = lbox
            appuifw.app.menu = []

        elif page == 1:
            #= Photo list
            lbox = None
            def p_edit_del():
                ix = lbox.current()
                sh.del_photo(ix)
                newlst = sh.get_photo_list()
                if ix >= len(newlst):
                    ix = len(newlst) - 1
                lbox.set_list(newlst, ix)
    
            def p_edit_add():
                global sh, ips_console
                #= Tabs must removed and added again due to dirBrowser
                appuifw.app.set_tabs([], lambda x: None)
                sel = pyS60uiutil.dirBrowser('Select photo').\
                      select(sh.config['photodir'])
                appuifw.app.set_tabs(tab_list, tabpages)
                appuifw.app.activate_tab(page)
                if sel == None: return
                sh.config['photodir'] = sel[0]
                ix = lbox.current()
                res = appuifw.popup_menu([u'After', u'Before'])
                if res == None: return
                if res == 0: ix += 1
                curr_app_info = pyS60uiutil.save_current_app_info()
                appuifw.app.body = ips_console.text
                disableCtrl = disableMenuAndExitkey()
                sh.add_photo(sel[0], sel[1], ix)
                disableCtrl.reactivate()
                pyS60uiutil.restore_app_info(curr_app_info)
                lbox.set_list(sh.get_photo_list(), ix)

            def _event(evt=None):
                #= View selected image here
                pass

            #lbox.bind(EKeyLeftArrow, lambda: self.__process_user_evt('back'))
            #lbox.bind(EKeyRightArrow, lambda: self.__process_user_evt('next'))

            lbox = appuifw.Listbox(sh.get_photo_list(), _event)
            appuifw.app.body = lbox
            appuifw.app.menu = [(u'Remove', p_edit_del), (u'Add', p_edit_add)]

        elif page == 2:
            #= Music list
            lbox = None
            def m_edit_del():
                ix = lbox.current()
                sh.del_music(ix)
                newlst = sh.get_music_list()
                if ix >= len(newlst):
                    ix = len(newlst) - 1
                lbox.set_list(newlst, ix)
    
            def m_edit_add():
                global sh
                #= Tabs must removed and added again due to dirBrowser
                appuifw.app.set_tabs([], lambda x: None)
                sel = pyS60uiutil.dirBrowser('Select music').\
                      select(sh.config['musicdir'])
                appuifw.app.set_tabs(tab_list, tabpages)
                appuifw.app.activate_tab(page)
                if sel == None: return
                sh.config['musicdir'] = sel[0]
                ix = lbox.current()
                res = appuifw.popup_menu([u'After', u'Before'])
                if res == None: return
                if res == 0: ix += 1
                sh.add_music(os.path.join(sel[0], sel[1]), ix)
                lbox.set_list(sh.get_music_list(), ix)

            def _event(evt=None):
                #= Play selected file here?
                pass

            #lbox.bind(EKeyLeftArrow, lambda: self.__process_user_evt('back'))
            #lbox.bind(EKeyRightArrow, lambda: self.__process_user_evt('next'))

            lbox = appuifw.Listbox(sh.get_music_list(), _event)
            appuifw.app.body = lbox
            appuifw.app.menu = [(u'Remove', m_edit_del), (u'Add', m_edit_add)]
            
            
    appuifw.app.exit_key_handler = edit_lock.signal
    appuifw.app.set_tabs(tab_list, tabpages)
    appuifw.app.activate_tab(0)
    tabpages(0)
    edit_lock.wait()
    appuifw.app.set_tabs([], lambda x: None)
    pyS60uiutil.restore_app_info(orgapp)
    sh.save()

def _selectShow():
    pre_l = len(SHOW_DIR_PREFIX)
    suf_l = len(SHOW_DIR_SUFFIX)
    list = []
    
    for item in os.listdir(IPS_DATA_DIRECTORY):
        if (item[0:pre_l] == SHOW_DIR_PREFIX) and \
               (item[0-suf_l:] == SHOW_DIR_SUFFIX):
            list.append(unicode(item[pre_l : 0-suf_l]))
    if len(list) == 0:
        appuifw.note(u'No shows available', 'error')
        return None
    sel = appuifw.popup_menu(list)
    if sel == None: return None
    return list[sel]

def loadShow():
    global sh
    show_title = _selectShow()
    if show_title == None: return
    sh.reset()
    sh.init_existing_show(show_title)
    if sh.load():
        appuifw.note(u'Photo show loaded successfully', 'conf')
        set_options_menu('showloaded')
    else:
        appuifw.note(u'Error while loading photo show', 'error')
        set_options_menu('initial')

def removeShow():
    global sh
    show_title = _selectShow()
    if show_title == None: return
    if appuifw.query(u'Want to remove ' + unicode(show_title), 'query'):
        if sh.delete(show_title) == True:
            set_options_menu('initial')

def volumeSettings():
    global sh
    if not sh.psound == None:
        vol = volumeControl(sh.psound, 'bar', None)
        sh.set_music_params(volume=vol)

def help():
    fv = pyS60uiutil.fileViewer('Help iPS', joystick=True)
    fv.load(os.path.join(os.path.split(appuifw.app.full_name())[0],
                         'ipshelp.txt'))
    fv.view()

def about():
    fv = pyS60uiutil.fileViewer('About iPS')
    fv.load(os.path.join(os.path.split(appuifw.app.full_name())[0],
                         'ipsabout.txt'))
    fv.view()

def set_options_menu(mode):
    if mode == 'initial':
        options_menu = [(u'Load',   loadShow),
                        (u'Create', createShow),
                        (u'Remove', removeShow),
                        (u'Help',   ((u'Help', help),
                                     (u'About', about)))]
    elif mode == 'showloaded':
        options_menu = [(u'Play',   playShow),
                        (u'Edit',   editShow),
                        (u'Load',   loadShow),
                        (u'Create', createShow),
                        (u'Remove', removeShow),
                        (u'Help',   ((u'Help', help),
                                     (u'About', about)))]
    elif mode == 'playing':
        options_menu = [(u'Volume', volumeSettings),
                        (u'Stop',   stopShow),
                        (u'Help',   ((u'Help', help),
                                     (u'About', about)))]
    appuifw.app.menu = options_menu

def stop_and_exit():
    global sh
    sh.stop()
    mylock.signal()

def photoShow():
    """
    The main program.
    """
    global sh, ips_console
    orgapp = pyS60uiutil.save_current_app_info()
    orgstderr = sys.stderr
    orgstdout = sys.stdout

    appuifw.app.title = u'instant Photo Show'
    ips_console = series60_console.Console()
    appuifw.app.body = ips_console.text
    sys.stderr = sys.stdout = ips_console
    
    sh = Show()
        
    appuifw.app.exit_key_handler = stop_and_exit
    set_options_menu('initial')
    mylock.wait()
    sh.save_parameters()
    sys.stderr = orgstderr
    sys.stdout = orgstdout
    pyS60uiutil.restore_app_info(orgapp)

if __name__ == '__main__':
    localpath = str(os.path.split(appuifw.app.full_name())[0])
    localpath = os.path.join(localpath, 'my')
    IPS_DATA_DIRECTORY = os.path.join(localpath, IPS_DATA_DIRECTORY)
    IPS_CONFIG_FILE = os.path.join(IPS_DATA_DIRECTORY, IPS_CONFIG_FILE)
    if not os.path.isdir(IPS_DATA_DIRECTORY): os.mkdir(IPS_DATA_DIRECTORY)
    photoShow()
